__name__ = "azskpy"
__version__ = "0.9.61"
PASSED = "Passed"
FAILED = "Failed"
MANUAL = "Verify"
ERROR = "Error"
HDI_METADATA_WRITE_PATH = "wasb:///AzSK_Metadata/Meta-{}.json"
HDI_SCANLOG_WRITE_PATH = "wasb:///AzSK_Logs/Scan-{}.csv"
LOG_ANALYTICS_API_VERSION = "2016-04-01"
